import lingam
from interpret.glassbox import ExplainableBoostingClassifier
from sklearn.metrics import accuracy_score


def timeout_handler(signum, frame):
    raise TimeoutError("Calculation timed out!")

DEFAULT_LOG_PATH = 'progress.txt'
DEFAULT_ERROR_LOG_PATH = 'progress_error.txt'


def preprocess_ds(ds):
    """
    Quick hack for incompatibility of classes between test and train set.
    Sometimes the population of a class is so small that it does not appear in both test and train data.
    Method just copies the lacking 1:1 - there should not be many of them
    Parameters
    ----------
    ds: Dataset

    Returns
    -------
    Fixed train and test set - both have same classes in
    """
    ds_df_train = ds.df_train
    ds_df_test = ds.df_test
    unique_classes_train = set(clas for clas in ds_df_train[ds.target])
    unique_classes_test = set(clas for clas in ds_df_test[ds.target])
    if unique_classes_test != unique_classes_train:
        has_train_but_not_test = unique_classes_train - unique_classes_test
        if len(has_train_but_not_test) > 0:
            filtered_rows = ds_df_train[ds_df_train[ds.target].isin(has_train_but_not_test)]
            ds_df_test = ds_df_test.append(filtered_rows)
        has_test_but_not_train = unique_classes_test - unique_classes_train
        if len(has_test_but_not_train):
            filtered_rows = ds_df_test[ds_df_test[ds.target].isin(has_test_but_not_train)]
            ds_df_train = ds_df_train.append(filtered_rows)
    return ds_df_train, ds_df_test

def train_ebm_model(ds, random_state, n_jobs=-1, interactions=0, debug=False):
    ds_df_train, ds_df_test = preprocess_ds(ds)
    model_clf = ExplainableBoostingClassifier(random_state=random_state, n_jobs=n_jobs, feature_types=ds.feature_types,
                                              interactions=interactions)
    model_clf.fit(ds_df_train[ds.features], ds_df_train[ds.target])

    y_pred = model_clf.predict(ds_df_test[ds.features])
    y_test = ds_df_test[ds.target]
    if debug:
        print(f"Model accuracy: {round(accuracy_score(y_test, y_pred), 2)}")

    return model_clf

def train_causal_model(dataset):
    causal_model = lingam.DirectLiNGAM()
    causal_model.fit(dataset.df_train[dataset.features])

    return causal_model




from .ccfs import generate_cfs, generate_cfs_bayesian
from .EBMCounterOptimizer import EBMCounterOptimizer
from .LogisticRegressionCounterOptimizer import LogisticRegressionCounterOptimizer
from .bayesian_model import train_bayesian_model, discretize_dataframe, discretize_ndarray
__all__ = [LogisticRegressionCounterOptimizer, EBMCounterOptimizer, generate_cfs, generate_cfs_bayesian,
           train_bayesian_model, discretize_dataframe, discretize_ndarray]


from .RandomSearchCounterOptimizer import RandomSearchCounterOptimizer
from .ACFX import ACFX
from .AcfxEBM import AcfxEBM
from .AcfxCustom import AcfxCustom
from .AcfxLinear import AcfxLinear

__all__ = [ACFX, AcfxLinear, AcfxEBM, AcfxCustom, RandomSearchCounterOptimizer]


import numpy as np
import pandas as pd
from overrides import overrides
from pgmpy.models import DiscreteBayesianNetwork
from sklearn.base import ClassifierMixin
from typing import Sequence, Tuple, Dict, Optional, List, Self
from .ACFX import ACFX
from .abstract import OptimizerType, ModelBasedCounterOptimizer


class AcfxCustom(ACFX):
    """
        AcfxCustom: A Counterfactual Explanation Model (using custom blackbox)
    """
    def __init__(self, blackbox: ClassifierMixin):
        """

        Parameters
        ----------
        blackbox:
            Custom blackbox explainer
        """
        super().__init__(blackbox)

    @overrides
    def counterfactual(self, query_instance: np.ndarray, desired_class: int, num_counterfactuals: int = 1, proximity_weight: float = 1,
                       sparsity_weight: float = 1, plausibility_weight: float = 0, diversity_weight: float = 1,
                       init_points: int = 10,
                       n_iter: int = 1000, sampling_from_model: bool = True) -> np.ndarray:

        if self.optimizer is None:
            raise ValueError("Optimizer must be initialized in fit() before calling counterfactual().")
        return super().counterfactual(query_instance, desired_class, num_counterfactuals, proximity_weight, sparsity_weight,
                                                plausibility_weight, diversity_weight, init_points,
                                      n_iter, sampling_from_model)

    def fit(self, X: pd.DataFrame, pbounds: Dict[str, Tuple[float, float]], optimizer : ModelBasedCounterOptimizer=None,
            causal_order: Optional[Sequence[int]]=None,
            adjacency_matrix: Optional[np.ndarray]=None, y=None, masked_features: Optional[List[str]] = None,
            categorical_indicator: Optional[List[bool]] = None, features_order: Optional[List[str]] = None,
            bayesian_causality:bool = False,  bayesian_model : Optional[DiscreteBayesianNetwork]=None, num_bins:Optional[int] = None) -> Self:
        """
        Fits explainer to the sampled data and blackbox model provided in the constructor

        :return:
        self
            Fitted estimator.

        Parameters
        ----------
        X : {sparse matrix} of shape (n_samples, n_features)
            Used for counterfactuals generation

        pbounds:
            The bounds for each feature to search over (dict with feature names as keys and tuple (min, max) as values).

        optimizer:
            Custom optimizer compliant with blackbox predictor

        causal_order:
            The order of variables in the causal graph.

        adjacency_matrix:
            The adjacency matrix representing the causal structure.

        y : array-like of shape (n_samples,)
            Target values used for blackbox model fitting only. You can provide fitted blackbox to constructor or fit it in this method by providing this parameter

        masked_features:
            List of interchangeable features

        categorical_indicator:
            True at the index where the variable should be treated as categorical

        features_order:
            order of features in query instance

        bayesian_causality:
            skip adjacency and calculate discrete bayesian network for causal loss.
            A discrete bayesian network will be fitted to calculate causal loss component.

        bayesian_model:
            optionally, provide pre-fitted discrete bayesian model, if bayesian_causality is True.
            If it is not provided and=bayesian_causality, it will be fitted with the num_bins param

        num_bins:
            Number of bins to use for discretizing continuous features

        """
        self.optimizer_type = OptimizerType.Custom
        if optimizer is None:
            raise ValueError("Optimizer must be given for AcfxCustom")
        self.optimizer = optimizer
        return super().fit(X, pbounds, causal_order, adjacency_matrix,
                    y, masked_features,categorical_indicator, features_order, bayesian_causality, bayesian_model, num_bins)


import numpy as np
from pgmpy.models import DiscreteBayesianNetwork
from scipy.spatial.distance import pdist, squareform

def compute_causal_penalty_bayesian_adjacency(samples, model : DiscreteBayesianNetwork):
    """
    Compute causal penalty for counterfactual samples based on Bayesian Network CPDs.

    Parameters:
    - samples: np.ndarray or pd.DataFrame (num_samples x num_features). All features must be categorical
    - model: fitted BayesianNetwork

    Returns:
    - penalty: float, average negative log-likelihood of samples under the model
    """
    penalty = 0.0
    n_samples = len(samples)
    nodes = list(model.nodes())
    for idx in range(n_samples):
        sample = samples[idx]
        sample_penalty = 0.0

        for node in nodes:
            parents = model.get_parents(node)
            if not parents:
                continue
            cpd = model.get_cpds(node)
            if not cpd:
                continue

            # Get parent values for this sample
            parent_values = [int(sample[nodes.index(p)]) for p in parents]
            node_value = sample[nodes.index(node)]

            # Find probability from CPD
            prob = cpd.get_value(**{node: int(node_value), **dict(zip(parents, parent_values))})
            if np.abs(prob) < 1e-8:
                prob = 1e-8
            sample_penalty += -np.log(prob)

        penalty += sample_penalty

    return penalty / n_samples

def compute_causal_penalty(samples, adjacency_matrix, sample_order, categorical=None, skip_categorical=True):
    """
    Calculate the inconsistency of samples with the given adjacency matrix.

    Parameters:
    - samples: np.ndarray, counterfactual samples to evaluate the causal penalty for
    - adjacency_matrix: np.ndarray, the adjacency matrix from DirectLiNGAM
    - samples: np.ndarray, the samples to evaluate (num_samples x num_features)
    - categorical: bool, categorical or continuous indicator
    - skip_categorical: bool, skip categorical indicator - when adjacency matrix is not categorical-feature effective.
    Even if the flag is true, we want the sample_order to be the same length as number of features in samples to keep the indexing
    Returns:
    - inconsistency: float, a measure of how inconsistent the samples are with the adjacency matrix
    """
    num_samples, num_features = samples.shape
    inconsistency = 0.0

    if categorical is None:
        categorical = [False] * len(sample_order)
    # Iterate through each feature and its causal parents
    included_to_loss = 0
    for i in sample_order:
        if skip_categorical and categorical[i]:
            continue
        included_to_loss += 1
        parents = np.where(adjacency_matrix[i, :] != 0)[0]
        if len(parents) > 0:
            # Predicted values based on parents
            predicted_values = np.dot(samples[:, parents], adjacency_matrix[i, parents])
            # Calculate the inconsistency as the mean squared error
            if categorical[i]:
                mse = (samples[:, i].round() != predicted_values.round()).astype(float)
            else:
                mse = (samples[:, i] - predicted_values) ** 2
            inconsistency += mse

    return np.sqrt(np.mean(inconsistency)) / max(1, included_to_loss)


def compute_yloss(model, cfs, desired_class):
    predicted_value = np.array(model.predict_proba(cfs))
    maxvalue = np.full((len(predicted_value)), -np.inf)
    for c in range(len(model.classes_)):
        if c != desired_class:
            maxvalue = np.maximum(maxvalue, predicted_value[:, c])
    yloss = np.maximum(0, maxvalue - predicted_value[:, int(desired_class)])
    return yloss,maxvalue


def compute_proximity_loss(cfs, query_instance, pbounds, features_order=None, feature_weights=None, categorical=None):
    """Compute weighted distance between two vectors."""
    # If feature_weights is None, assign an array of ones with the size of cfs.shape[1]

    if feature_weights is None:
        feature_weights = np.ones(cfs.shape[1])
    if categorical is None:
        categorical = [False] * cfs.shape[1]

    diff = abs(cfs - query_instance)
    diff = [1 if categorical[i] and v > 0 else v for i, v in enumerate(diff[0, :])]

    diff = [
        v / (pbounds[features_order[i]][1] - pbounds[features_order[i]][0]) if not categorical[i] and features_order[
            i] in pbounds.keys() else v for i, v in enumerate(diff)]
    product = np.multiply(
        (diff),
        feature_weights)
    product = product.reshape(-1, product.shape[-1])
    proximity_loss = np.sum(product, axis=1)

    # Dividing by the sum of feature weights to normalize proximity loss
    return proximity_loss / sum(feature_weights)

def compute_sparsity_loss(cfs, query_instance):
    """Compute weighted distance between two vectors."""
    sparsity_loss = np.count_nonzero(cfs - query_instance, axis=1)
    return sparsity_loss / cfs.shape[1]  # Dividing by the number of features to normalize sparsity loss


def compute_diversity_loss(cfs, low=1e-6, high=1e-5):
    # Compute pairwise distances
    pairwise_distances = pdist(cfs)

    # Convert the distances to a square matrix form
    # TODO: gower distance?
    distance_matrix = squareform(pairwise_distances)

    perturbations = np.random.uniform(low=low, high=high, size=distance_matrix.shape[0])

    # Add the random perturbations to the diagonal elements of the transformed matrix
    np.fill_diagonal(distance_matrix, distance_matrix.diagonal() + perturbations)

    transformed_matrix = 1 / (1 + distance_matrix)

    # Calculate the determinant of the transformed matrix
    determinant = np.linalg.det(transformed_matrix)
    return determinant

def _compute_loss_common(
    yloss, conditional_term,
    proximity_loss, sparsity_loss, plausibility_loss, diversity_loss,
    proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
    cfs
):
    """Helper to compute final loss array."""
    total_loss = yloss + conditional_term * (
        (diversity_loss * diversity_weight) +
        (proximity_weight * proximity_loss) +
        (sparsity_weight * sparsity_loss) +
        (plausibility_loss * plausibility_weight)
    )
    loss = np.reshape(np.array(total_loss), (-1, 1))
    index = np.reshape(np.arange(len(cfs)), (-1, 1))
    return np.concatenate([index, loss], axis=1)


def compute_loss_generic(
    model, cfs, query_instance, desired_class,
    proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
    pbounds, features_order, categorical=None, allcfs=None,
    plausibility_fn=None
):
    """Generic loss computation with injected plausibility function."""
    if allcfs is None:
        allcfs = []

    yloss, _ = compute_yloss(model, cfs, desired_class)
    conditional_term = 1.0 / (
        yloss + 1 + sum([proximity_weight, sparsity_weight, plausibility_weight, diversity_weight])
    )

    proximity_loss = compute_proximity_loss(
        cfs, query_instance, pbounds, features_order=features_order, categorical=categorical
    ) if proximity_weight > 0 else 0.0

    sparsity_loss = compute_sparsity_loss(cfs, query_instance) if sparsity_weight > 0 else 0.0

    plausibility_loss = plausibility_fn() if plausibility_weight > 0 else 0.0

    diversity_loss = 1 - np.abs(compute_diversity_loss(allcfs)) if len(allcfs) > 1 and diversity_weight > 0 else 0.0

    return _compute_loss_common(
        yloss, conditional_term,
        proximity_loss, sparsity_loss, plausibility_loss, diversity_loss,
        proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
        cfs
    )

def compute_loss_bayesian(model, cfs, cfs_categorized, query_instance, desired_class, bayesian_model : DiscreteBayesianNetwork,
                          proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
                          pbounds, features_order, masked_features, categorical=None, allcfs=None):
    return compute_loss_generic(
        model, cfs, query_instance, desired_class,
        proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
        pbounds, features_order, categorical, allcfs,
        plausibility_fn=lambda: compute_causal_penalty_bayesian_adjacency(cfs_categorized, bayesian_model)
    )

def compute_loss(model, cfs, query_instance, desired_class, adjency_matrix, causal_order,
                 proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
                 pbounds, features_order, masked_features, categorical=None, allcfs=None):
    return compute_loss_generic(
        model, cfs, query_instance, desired_class,
        proximity_weight, sparsity_weight, plausibility_weight, diversity_weight,
        pbounds, features_order, categorical, allcfs,
        plausibility_fn=lambda: compute_causal_penalty(cfs, adjency_matrix, causal_order, categorical=categorical)
    )




from typing import Sequence, Tuple, Dict, Optional, List, Self

import numpy as np
import pandas as pd
from overrides import overrides
from pgmpy.models import DiscreteBayesianNetwork
from sklearn.linear_model._base import LinearClassifierMixin

from .ACFX import ACFX
from .abstract import OptimizerType


class AcfxLinear(ACFX):
    """
        AcfxCustom: A Counterfactual Explanation Model (using linear additive model as blackbox)
    """
    def __init__(self, blackbox: LinearClassifierMixin):
        """

        Parameters
        ----------
        blackbox:
            Linear blackbox explainer
        """
        super().__init__(blackbox)

    @overrides
    def fit(self, X:pd.DataFrame, pbounds:Dict[str, Tuple[float, float]], causal_order:Optional[Sequence[int]]=None,
            adjacency_matrix:Optional[np.ndarray]=None,
            y=None, masked_features:Optional[List[str]] = None,
            categorical_indicator:Optional[List[bool]] =None, features_order:Optional[List[str]] =None,
            bayesian_causality:bool = False,  bayesian_model : Optional[DiscreteBayesianNetwork]=None,
            num_bins:Optional[int] = None) -> Self:
        self.optimizer_type = OptimizerType.LinearAdditive
        return super().fit(X, pbounds, causal_order, adjacency_matrix,
                    y, masked_features,categorical_indicator, features_order, bayesian_causality, bayesian_model, num_bins)



from typing import Dict, Tuple

import numpy as np
import pandas as pd
from .abstract import ModelBasedCounterOptimizer
from overrides import overrides


class RandomSearchCounterOptimizer(ModelBasedCounterOptimizer):
    def __init__(self, model, X: pd.DataFrame, feature_bounds: Dict[str, Tuple[float, float]], n_iter: int = 100):
        if not hasattr(model, 'predict_proba'):
            raise AttributeError("Model must implement predict_proba()")
        self.model = model
        self.X = X
        self.feature_bounds = feature_bounds
        self.n_iter = n_iter

    @overrides
    def optimize_proba(self, target_class: int, feature_masked: list[str]) -> Dict[str, float]:
        base_instance = self.X.mean().copy()
        best_instance = base_instance.copy()
        best_score = self.model.predict_proba([base_instance])[0][target_class]

        for _ in range(self.n_iter):
            candidate = base_instance.copy()
            for feature_name in self.X.columns:
                if feature_name in feature_masked and feature_name in self.feature_bounds:
                    min_val, max_val = self.feature_bounds[feature_name]
                    candidate[feature_name] = np.random.uniform(min_val, max_val)

            score = self.model.predict_proba([candidate])[0][target_class]
            if score > best_score:
                best_score = score
                best_instance = candidate.copy()

        return best_instance.to_dict()


from typing import Sequence, Tuple, Dict, Optional, List, Self

import numpy as np
import pandas as pd
from pgmpy.models import DiscreteBayesianNetwork
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.base import ClassifierMixin
from abc import ABC, abstractmethod

from .evaluation import generate_cfs_bayesian
from .abstract import OptimizerType
from .evaluation import generate_cfs
from .evaluation import train_bayesian_model

class ACFX(ABC, BaseEstimator, TransformerMixin):
    """
        ACFX: A Counterfactual Explanation Model
    """
    def __init__(self, blackbox:ClassifierMixin):
        """

        Parameters
        ----------
        blackbox:
            Blackbox explainer
        """
        self.bayesian_causality = None
        self.num_bins = None
        self.blackbox = blackbox
        self.optimizer = None
        self.optimizer_type = None
        self.X = None
        self.categorical_indicator = None
        self.features_order = None
        self.pbounds = None
        self.adjacency_matrix = None
        self.causal_order = None
        self.masked_features = None
        self.bayesian_model = None


    @abstractmethod
    def fit(self, X: pd.DataFrame, pbounds: Dict[str, Tuple[float, float]], causal_order: Optional[Sequence[int]]=None,
            adjacency_matrix: Optional[np.ndarray]=None, y=None, masked_features: Optional[List[str]] = None,
            categorical_indicator: Optional[List[bool]] = None, features_order: Optional[List[str]] = None,
            bayesian_causality:bool = False, bayesian_model : Optional[DiscreteBayesianNetwork]=None, num_bins:Optional[int] = None) -> Self:
        """
        Fits explainer to the sampled data and blackbox model provided in the constructor

        :return:
        self
            Fitted estimator.

        Parameters
        ----------
        X : {sparse matrix} of shape (n_samples, n_features)
            Used for counterfactuals generation

        pbounds:
            The bounds for each feature to search over (dict with feature names as keys and tuple (min, max) as values).

        causal_order:
            The order of variables in the causal graph.

        adjacency_matrix:
            The adjacency matrix representing the causal structure.

        y : array-like of shape (n_samples,).
            Target values used for blackbox model fitting only. You can provide fitted blackbox to constructor or fit it in this method by providing this parameter

        masked_features:
            List of interchangeable features

        categorical_indicator:
            True at the index where the variable should be treated as categorical

        features_order:
            order of features in query instance

        bayesian_causality:
            skip adjacency and calculate discrete bayesian network for causal loss.
            A discrete bayesian network will be fitted to calculate causal loss component.

        bayesian_model:
            optionally, provide pre-fitted discrete bayesian model, if bayesian_causality is True.
            If it is not provided and=bayesian_causality, it will be fitted with the num_bins param

        num_bins:
            Number of bins to use for discretizing continuous features (bayesian causality only)

        """
        if y is not None:
            self.blackbox.fit(X, y)
        self.X = X
        self.categorical_indicator = categorical_indicator
        self.features_order = features_order
        self.bayesian_causality = bayesian_causality

        if not self.bayesian_causality:
            self.adjacency_matrix = adjacency_matrix
            self.causal_order = causal_order
        else:
            self.num_bins = num_bins
            if bayesian_model is None:
                self.bayesian_model = train_bayesian_model(self.X, self.categorical_indicator, self.num_bins)
            else:
                self.bayesian_model = bayesian_model
        self.pbounds = pbounds
        self.masked_features = masked_features
        return self

    def predict(self, X):
        """
        Predicts using blackbox model

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Used for counterfactuals generation

        Returns
        -------
        Prediction class labels for samples in X by blackbox model
        """
        return self.blackbox.predict(X)

    def counterfactual(self, query_instance: np.ndarray, desired_class:int, num_counterfactuals: int =1, proximity_weight : float =1,
                       sparsity_weight : float =1, plausibility_weight : float =0, diversity_weight : float =1, init_points : int =10,
                       n_iter : int =1000, sampling_from_model : bool=True) -> np.ndarray:
        """
        Generates counterfactuals

        Parameters
        ----------
        query_instance:
            The instance to generate counterfactuals for.

        desired_class:
            The target class for the counterfactuals.

        num_counterfactuals:
            The number of counterfactual instances to generate.

        proximity_weight:
            Weight for proximity loss component

        sparsity_weight:
            Weight for sparsity loss component

        plausibility_weight:
            Weight for plausibility loss component

        diversity_weight:
            Weight for diversity loss component

        init_points:
            Number of initial points for Bayesian Optimization.

        n_iter:
            Number of iterations for Bayesian Optimization.

        sampling_from_model:
            true if you want to generate samples from model after sampling from data and generating with relationship graph

        Returns
        -------
        np.ndarray:
            The generated counterfactuals that minimize the loss function.
        """
        if plausibility_weight > 0:
            if not self.bayesian_causality:
                if self.causal_order is None:
                    raise ValueError("Causal order must be provided if plausibility loss is on")
                if self.adjacency_matrix is None:
                    raise ValueError("adjacency_matrix must be provided")
                if self.adjacency_matrix.shape[0] != self.adjacency_matrix.shape[1]:
                    raise ValueError("adjacency matrix must have same number of rows and columns")
                if self.adjacency_matrix.shape[0] != len(self.causal_order):
                    raise ValueError("adjacency matrix must be of same length as causal order")
            else:
                if self.bayesian_model is None:
                    raise ValueError("Bayesian model must be provided")
                if self.num_bins is None:
                    raise ValueError("num_bins must be provided")

        if query_instance is None:
            raise ValueError("query_instance must not be None")
        if self.optimizer_type is None:
            raise ValueError("optimizer_type must be set via fit() before calling counterfactual()")
        if self.optimizer is None and self.optimizer_type is OptimizerType.Custom:
            raise ValueError("optimizer must be set before calling counterfactual()")
        if self.optimizer_type is OptimizerType.LinearAdditive:
            if not hasattr(self.blackbox, 'coef_'):
                raise AttributeError('optimizer requires model.coef_ as linear coefficients to be set')
        if self.bayesian_model is None:
            return generate_cfs(query_instance=query_instance,
                                desired_class=desired_class,
                                adjacency_matrix=self.adjacency_matrix,
                                causal_order=self.causal_order,
                                proximity_weight=proximity_weight,
                                sparsity_weight=sparsity_weight,
                                plausibility_weight=plausibility_weight,
                                diversity_weight=diversity_weight,
                                bounds=self.pbounds,
                                model=self.blackbox,
                                features_order=self.features_order,
                                masked_features= self.masked_features,
                                categorical_indicator= self.categorical_indicator,
                                X=self.X,
                                num_cfs=num_counterfactuals,
                                init_points=init_points,
                                n_iter=n_iter,
                                sampling_from_model=sampling_from_model,
                                optimizer_type=self.optimizer_type,
                                optimizer=self.optimizer)
        else:
            return generate_cfs_bayesian(query_instance=query_instance,
                                desired_class=desired_class,
                                bayesian_model=self.bayesian_model,
                                categorical_bins_num=self.num_bins,
                                proximity_weight=proximity_weight,
                                sparsity_weight=sparsity_weight,
                                plausibility_weight=plausibility_weight,
                                diversity_weight=diversity_weight,
                                bounds=self.pbounds,
                                model=self.blackbox,
                                features_order=self.features_order,
                                masked_features= self.masked_features,
                                categorical_indicator= self.categorical_indicator,
                                X=self.X,
                                num_cfs=num_counterfactuals,
                                init_points=init_points,
                                n_iter=n_iter,
                                sampling_from_model=sampling_from_model,
                                optimizer_type=self.optimizer_type,
                                optimizer=self.optimizer)






from typing import List, Dict

import pandas as pd
from interpret.glassbox import ExplainableBoostingClassifier
from overrides import overrides
from sklearn.utils.extmath import softmax
import numpy as np

from ..abstract import ModelBasedCounterOptimizer


class EBMCounterOptimizer(ModelBasedCounterOptimizer):
    # Y_TEST = '_eco_y_test'
    # Y_PRED = '_eco_y_pred'
    # IS_MODIFIABLE = '_eco_is_modifiable'

    def __init__(self, model: ExplainableBoostingClassifier, X: pd.DataFrame):
        self.model = model
        self.X = X
        self.updated_features = {}

    def _get_optimized_feature_value(self, feature_name, feature_idx, feature_val, features, feature_masked, term_idx,
                                     class_idx) -> Dict[str, float]:
        """
        Returns
        -------
        feature value with maximum score for given target class.

        @Todo Needs changes to return optimized value due to given strategy.
        """
        if feature_name in feature_masked and feature_name not in self.updated_features:
            if len(self.model.term_scores_[term_idx].shape) > 1:
                class_term_scores = self.model.term_scores_[term_idx].T[class_idx]
            else:
                class_term_scores = self.model.term_scores_[term_idx] if class_idx == 1 else 1 - self.model.term_scores_[
                    term_idx]
            class_max = np.max(class_term_scores)
            try:
                feature_score_idx = np.where(class_term_scores[1:-1] == class_max)[0][0]  ##this is score, not value imho
            except:
                print(np.where(class_term_scores[1:-1] == class_max))
            # we bin differently for main effects and pairs, so first
            # get the list containing the bins for different resolutions
            bin_levels = self.model.bins_[feature_idx]
            # print(f'Feature score index for feature {feature_name} is {feature_score_idx} which represents score equal: {class_max} test: {class_term_scores[feature_score_idx+1]}')
            # what resolution do we need for this term (main resolution, pair
            # resolution, etc.), but limit to the last resolution available
            bins = bin_levels[min(len(bin_levels), len(features)) - 1]

            if len(bins) == 0:
                feature_val = self.X[feature_name].sample(1).values[0]
            else:
                if isinstance(bins, dict):
                    # categorical feature
                    # 'unknown' category strings are in the last bin (-1)
                    feature_val = list(bins.values())[
                        feature_score_idx - 1]  # if maxscore was 0, or -1 just assign random value
                else:
                    # continuous feature
                    # Get the lower and upper bounds of the specified bin
                    lower_idx = feature_score_idx - 1
                    upper_idx = feature_score_idx

                    if lower_idx == -1:
                        lower = self.model.feature_bounds_[feature_idx][0]
                    else:
                        lower = bins[lower_idx]

                    if upper_idx == len(bins):
                        upper = self.model.feature_bounds_[feature_idx][1]
                    else:
                        upper = bins[upper_idx]
                    # print(f'Drawing randomly from :{lower} to {upper}')

                    # Draw a random number from the range defined by the bin
                    feature_val = np.random.uniform(lower, upper)

            # print(f'This translates into feature value: {feature_val}')

            self.updated_features.update({feature_name: feature_val})
        elif feature_name in self.updated_features:
            feature_val = self.updated_features.get(feature_name)
        else:
            self.updated_features.update({feature_name: feature_val})

        return feature_val

    @overrides
    def optimize_proba(self, target_class : int, feature_masked: List[str]) -> Dict[str, float]:
        """
        The method calculates probabilities taking into account the optimization of given parameters towards the target class.
        Method is based on a default ebm's predict_proba

        Parameters:
        ebm:
            Trained EBM model
        X:
            Dataset
        target_class:
            Target class from which we take the features
        featured_masked:
            List of interchangeable features
        """
        if target_class not in self.model.classes_:
            raise KeyError(f'Class "{target_class}" does not exists in given EBM model')

        class_idx = np.where(self.model.classes_ == target_class)[0][0]
        self.updated_features = {}
        sample_scores = []
        cf = {}
        for index, sample in self.X.iterrows():
            # start from the intercept for each sample
            score = self.model.intercept_.copy()
            if isinstance(score, float) or len(score) == 1:
                # regression or binary classification
                score = float(score)

            # we have 2 terms, so add their score contributions
            for term_idx, features in enumerate(self.model.term_features_):
                term_scores = self.model.term_scores_[term_idx]
                # indexing into a tensor requires a multi-dimensional index
                tensor_index = []
                # main effects will have 1 feature, and pairs will have 2 features
                for feature_idx in features:
                    feature_name = self.model.feature_names_in_[feature_idx]  # Get the feature name by index
                    feature_val = sample[feature_name]  # Use the feature name to get the correct value from the sample
                    bin_idx = 0  # if missing value, use bin index 0

                    if feature_val is not None and feature_val is not np.nan:
                        # we bin differently for main effects and pairs, so first
                        # get the list containing the bins for different resolutions
                        bin_levels = self.model.bins_[feature_idx]

                        # what resolution do we need for this term (main resolution, pair
                        # resolution, etc.), but limit to the last resolution available
                        bins = bin_levels[min(len(bin_levels), len(features)) - 1]

                        # here is where the magic is located
                        feature_val = self._get_optimized_feature_value(feature_name, feature_idx, feature_val,
                                                                        features, feature_masked, term_idx, class_idx)

                        if isinstance(bins, dict):
                            # categorical feature
                            # 'unknown' category strings are in the last bin (-1)
                            bin_idx = bins.get(feature_val, -1)
                            if bin_idx == -1:
                                # check value as string
                                bin_idx = bins.get(str(feature_val), -1)
                        else:
                            # continuous feature
                            try:
                                # try converting to a float, if that fails it's 'unknown'
                                feature_val = float(feature_val)
                                # add 1 because the 0th bin is reserved for 'missing'
                                bin_idx = np.digitize(feature_val, bins) + 1
                            except ValueError:
                                # non-floats are 'unknown', which is in the last bin (-1)
                                bin_idx = -1

                        if len(term_scores.shape) > 1:
                            if bin_idx >= term_scores.T.shape[1]:
                                bin_idx = -1
                            sc = term_scores.T[class_idx][bin_idx]
                        else:
                            if bin_idx >= term_scores.shape[0]:
                                bin_idx = -1
                            sc = term_scores[bin_idx]
                        # print(f'And feature value {feature_val} translates back to bin index: {bin_idx} which represents score: {sc}')

                        tensor_index.append(bin_idx)

                # local_score is also the local feature importance
                local_score = term_scores[tuple(tensor_index)]

                score += local_score
            sample_scores.append(score)

        predictions = np.array(sample_scores)

        if hasattr(self.model, 'classes_'):
            # classification
            if len(self.model.classes_) == 2:
                # binary classification

                # softmax expects two logits for binary classification
                # the first logit is always equivalent to 0 for binary classification
                predictions = [[0, x] for x in predictions]
            predictions = softmax(predictions)

        #return predictions, self.updated_features
        return self.updated_features

    # def check_samples(self, target_class, y_test_key, feature_masked):
    #     X = self.X.copy()
    #     predictions = self.optimize_proba(target_class, feature_masked)
    #     X.loc[:, self.Y_TEST] = np.argmax(predictions, axis=1)
    #     X.loc[:, self.Y_PRED] = X[self.Y_TEST].map({key: val for key, val in enumerate(self.model.classes_)})
    #     X.loc[:, self.IS_MODIFIABLE] = np.where(X[y_test_key] != X[self.Y_PRED], 1, 0)
    #
    #     return X


import random
from typing import Sequence, Dict, Tuple, Optional, List
import traceback
import numpy as np
import optuna
import pandas as pd

from interpret.glassbox import ExplainableBoostingClassifier
from pgmpy.models import DiscreteBayesianNetwork
from sklearn.base import ClassifierMixin
from sklearn.linear_model._base import LinearClassifierMixin
from ..abstract import ModelBasedCounterOptimizer
from ..abstract import OptimizerType
from .LogisticRegressionCounterOptimizer import LogisticRegressionCounterOptimizer
from .loss import compute_loss, compute_loss_bayesian
from .EBMCounterOptimizer import EBMCounterOptimizer
from .bayesian_model import discretize_ndarray

def __generate_single_cf(query_instance, desired_class, adjacency_matrix, causal_order, proximity_weight, sparsity_weight,
                         plausibility_weight,
                         diversity_weight, bounds, model, optimizer:ModelBasedCounterOptimizer, sampling_from_model:bool,
                         categorical_indicator=None, features_order=None, masked_features=None, cfs=[], X=None,
                         init_points=5, n_iter=1000, bayesian_causality : bool = False, bins_num: Optional[int] = None,
                         bayesian_model: Optional[DiscreteBayesianNetwork] = None):
    """
    Generate a single counterfactual that minimizes the loss function using Optuna.
    """
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    def update_masked_features_dict(features_order, masked_features, **params):
        for feature, value in zip(features_order, query_instance):
            if feature not in masked_features:
                params[feature] = value
        return params

    def black_box_function(trial):
        # Suggest values for each feature based on bounds
        kwargs = {}
        for i, key in enumerate(features_order):
            if categorical_indicator[i]:
                kwargs[key] = trial.suggest_categorical(key, list(range(int(bounds[key][0]), int(bounds[key][1]) + 1)))
            else:
                kwargs[key] = trial.suggest_float(key, bounds[key][0], bounds[key][1], log=False)

        kwargs = update_masked_features_dict(features_order, masked_features, **kwargs)
        # todo rounding
        cf = np.array([[int(round(kwargs[key])) if categorical_indicator[i] else kwargs[key]
                        for i, key in enumerate(features_order)]]).reshape(1, -1)

        if len(cfs) > 0:
            cfst = np.vstack(cfs + [cf])
        else:
            cfst = cf

        if not bayesian_causality:
            loss = compute_loss(model,
                                cf,
                                query_instance,
                                desired_class,
                                adjacency_matrix,
                                causal_order,
                                proximity_weight,
                                sparsity_weight,
                                plausibility_weight,
                                diversity_weight,
                                pbounds=bounds,
                                features_order=features_order,
                                masked_features=masked_features,
                                categorical=categorical_indicator,
                                allcfs=cfst)
        else:
            categorized = discretize_ndarray(cf, categorical_indicator, bins_num)
            loss = compute_loss_bayesian(model, cf, categorized, query_instance, desired_class, bayesian_model,
                                         proximity_weight, sparsity_weight, plausibility_weight,
                                diversity_weight, pbounds=bounds, features_order=features_order,
                                masked_features=masked_features,
                                categorical=categorical_indicator, allcfs=cfst)
        # Return the first value in the loss array, Optuna needs a single scalar value to minimize
        return -loss[0, 1]

    # Initialize Optuna study
    study = optuna.create_study(direction='maximize')

    # Define seen_points set to track uniqueness of points
    seen_points = set()

    def is_unique_point(point_dict):
        point_tuple = tuple(point_dict[key] for key in features_order)
        if point_tuple in seen_points:
            return False
        seen_points.add(point_tuple)
        return True

    def clip_values(cf, pbounds):
        # Create a new dictionary with clipped values based on the bounds in pbounds
        return {feature: np.clip(value, pbounds[feature][0], pbounds[feature][1])
                for feature, value in cf.items()}


    def sample_from_data(Xdesired, features_order, init_points, sampled_trials, study):
        sample_size = min(int(init_points * 0.5), len(Xdesired))
        Xsample = Xdesired.sample(sample_size)
        for i, r in Xsample.iterrows():
            candidate_point = dict(r[features_order])
            if is_unique_point(candidate_point):
                trial_params = {key: candidate_point[key] for key in features_order}
                study.enqueue_trial(trial_params)
                sampled_trials += 1
        init_points = max(0, init_points - sampled_trials)
        return init_points, sample_size, sampled_trials

    def sample_from_model(Xdesired, bounds, desired_class, features_order, init_points,
                          masked_features, sample_size, sampled_trials, study):
        def optimizer_iteration(masked_features, total_lists: list, optimizer: ModelBasedCounterOptimizer,
                                desired_class):
            num_elements = random.randint(1, len(masked_features))
            selected_features = random.sample(masked_features, num_elements)
            set_to_check = set(selected_features)
            # Check if the set is in the list of sets
            found = any(set(item) == set_to_check for item in total_lists)
            if found:
                return None
            total_lists.append(selected_features)
            try:
                cf = optimizer.optimize_proba(desired_class, feature_masked=selected_features)
                return cf
            except Exception as ex:
                print(f'optimize_proba error occured: {ex}')
                print(traceback.format_exc())
                return None
        print(
            f'Sampling from model... Already sampled {sample_size} from {Xdesired.shape[0]} possible in data sampling...')
        # optimizer = get_ebm_optimizer(model, pd.DataFrame(query_instance.reshape(1, -1), columns=features_order))
        total_lists = []
        for i in range(min(init_points, 2 ** len(masked_features))):
            cf = optimizer_iteration(masked_features, total_lists, optimizer, desired_class)
            if cf is None:
                continue

            # check if values of cf are aligned with the ranges, and if not, clip it to the range
            cf = clip_values(cf, bounds)

            if isinstance(cf, dict):
                cf_dict = cf
            elif isinstance(cf, np.ndarray) or isinstance(cf, list):
                cf_dict = {key: cf[i] for i, key in enumerate(features_order)}
            else:
                raise ValueError("Unexpected format for cf")

            if is_unique_point(cf_dict):
                sampled_trials += 1
                study.enqueue_trial(cf_dict)
        return sampled_trials

    sampled_trials = 0
    if X is not None:
        Xdesired = X[model.predict(X) == desired_class].drop_duplicates()
        init_points, sample_size, sampled_trials = sample_from_data(X, features_order,
                                                                              init_points,
                                                                              sampled_trials, study)
        if init_points > 0 and sampling_from_model:
            sampled_trials = sample_from_model(Xdesired, bounds, desired_class, features_order,
                                               init_points, masked_features, sample_size, sampled_trials,
                                               study)
    else:
        raise ValueError('X is None')

    # Optimize the study with the defined number of iterations
    study.optimize(black_box_function, n_trials=n_iter + sampled_trials)

    # Extract the best parameters (counterfactual)
    best_params = study.best_params
    best_params = update_masked_features_dict(features_order, masked_features, **best_params)
    # todo rounding should be done to the featoure boundaries, not to nearest integer
    best_cf = np.array([[int(round(best_params[key])) if categorical_indicator[i] else best_params[key]
                         for i, key in enumerate(features_order)]]).reshape(1, -1)

    return best_cf

def _generate_single_cf(query_instance, desired_class, adjacency_matrix, causal_order, proximity_weight, sparsity_weight,
                        plausibility_weight,
                        diversity_weight, bounds, model, categorical_indicator=None, features_order=None,
                        masked_features=None, cfs=[], X=None, init_points=5, n_iter=1000,
                        optimizer_type: OptimizerType = OptimizerType.EBM, optimizer=None, sampling_from_model=True,
                        bayesian_causality : bool = False, bins_num: Optional[int] = None, bayesian_model: Optional[DiscreteBayesianNetwork] = None):

    def is_null_or_empty(var):
        if var is None:
            return True
        if isinstance(var, (list, tuple, dict, set)) and len(var) == 0:
            return True
        if isinstance(var, np.ndarray) and var.size == 0:
            return True
        return False

    def get_optimizer():
        def get_ebm_optimizer(model_classifier: ExplainableBoostingClassifier, query_instance: pd.DataFrame):
            return EBMCounterOptimizer(model_classifier, query_instance)
        def get_logistic_regression_optimizer(model_classifier: LinearClassifierMixin, query_instance: pd.DataFrame):
            return LogisticRegressionCounterOptimizer(model_classifier, query_instance, bounds)

        if optimizer_type == OptimizerType.EBM:
            return get_ebm_optimizer(model, pd.DataFrame(query_instance.reshape(1, -1), columns=features_order))
        if optimizer_type == OptimizerType.LinearAdditive:
            lro = get_logistic_regression_optimizer(model,
                                                    pd.DataFrame(query_instance.reshape(1, -1), columns=features_order))
            return lro
        if optimizer_type == OptimizerType.Custom:
            return optimizer
        else:
            raise NotImplementedError()

    if bayesian_causality:
        if bayesian_model is None:
            raise ValueError("bayesian_model must be provided for bayesian causal loss mode")
        if bins_num is None:
            raise ValueError("bins_num must be provided for bayesian causal loss mode")
    if is_null_or_empty(categorical_indicator):
        categorical_indicator = [False] * len(bounds)
    if is_null_or_empty(features_order):
        features_order = list(bounds.keys())
    if is_null_or_empty(masked_features):
        masked_features = features_order
    return __generate_single_cf(query_instance=query_instance, desired_class=desired_class, adjacency_matrix=adjacency_matrix,
                                causal_order=causal_order, proximity_weight=proximity_weight, sparsity_weight=sparsity_weight,
                                plausibility_weight=plausibility_weight, diversity_weight=diversity_weight, bounds=bounds,
                                model=model, optimizer=get_optimizer(), sampling_from_model= sampling_from_model,
                                categorical_indicator=categorical_indicator, features_order= features_order,
                                masked_features=masked_features, cfs=cfs, X= X, init_points= init_points, n_iter=n_iter, bayesian_causality= bayesian_causality,
                                bins_num=bins_num, bayesian_model = bayesian_model)


def generate_cfs_bayesian(query_instance:np.ndarray, desired_class:int, bayesian_model: DiscreteBayesianNetwork, categorical_bins_num: int,
                          proximity_weight : float, sparsity_weight: float, plausibility_weight: float, diversity_weight: float,
                          bounds:Dict[str, Tuple[float, float]], model:ClassifierMixin, features_order:Optional[List[str]] =None,
                 masked_features:Optional[List[str]] =None, categorical_indicator:Optional[List[bool]] =None, X:Optional[pd.DataFrame] =None,
                 num_cfs:int=1, init_points:int=10, n_iter:int=1000, sampling_from_model:bool=False,
                 optimizer_type : OptimizerType = OptimizerType.EBM, optimizer : ModelBasedCounterOptimizer=None)-> np.ndarray:
    """
    Generate multiple counterfactuals that minimize the loss function using Bayesian Optimization and bayesian network to calculate causal loss.

    Parameters:
        query_instance:
            The instance to generate counterfactuals for.
        desired_class:
            The target class for the counterfactuals.
        bayesian_model:
            Fitted bayesian network
        categorical_bins_num:
            Number of bins that were used to discretize continuous features.
        proximity_weight:
            Weight for proximity loss component
        sparsity_weight:
            Weight for sparsity loss component
        plausibility_weight:
            Weight for plausibility loss component
        diversity_weight:
            Weight for diversity loss component
        bounds:
            The bounds for each feature to search over (dict with feature names as keys and tuple (min, max) as values).
        model:
            The predictive model used to predict class labels.
        features_order:
            order of features in query instance
        masked_features:
            masked features vector (features to skip)
        categorical_indicator:
            True at the index where the variable should be treated as categorical
        num_cfs:
            The number of counterfactual instances to generate.
        init_points:
            Number of initial points for Bayesian Optimization.
        n_iter:
            Number of iterations for Bayesian Optimization.
        X:
            training dataset to sample from.
        sampling_from_model:
            true if you want to generate samples from model after sampling from data and generating with relationship graph
        optimizer_type:
            type of optimizer used on model to generate counterfactuals. If you use OptimizerType.Custom, you need to provide optimizer instance
        optimizer:
            instance of optimizer (use for optimizer_type = OptimizerType.Custom)

    Returns
    -------
    np.ndarray:
        The generated counterfactuals that minimize the loss function.
    """
    cfs = []
    for _ in range(num_cfs):
        cf = _generate_single_cf(query_instance, desired_class, None,
                                 None, proximity_weight, sparsity_weight,
                                 plausibility_weight, diversity_weight,
                                 bounds, model, categorical_indicator, features_order, masked_features=masked_features,
                                 cfs=cfs, X=X, init_points=init_points, n_iter=n_iter,
                                 sampling_from_model=sampling_from_model, optimizer_type=optimizer_type,
                                 optimizer=optimizer, bayesian_causality=True, bins_num=categorical_bins_num, bayesian_model=bayesian_model)
        cfs.append(cf)

    return np.vstack(cfs)

def generate_cfs(query_instance:np.ndarray, desired_class:int, adjacency_matrix:np.ndarray, causal_order : Sequence[int], proximity_weight : float,
                 sparsity_weight: float, plausibility_weight: float, diversity_weight: float, bounds:Dict[str, Tuple[float, float]],
                 model:ClassifierMixin, features_order:Optional[List[str]] =None,
                 masked_features:Optional[List[str]] =None, categorical_indicator:Optional[List[bool]] =None, X:Optional[pd.DataFrame] =None,
                 num_cfs:int=1, init_points:int=10, n_iter:int=1000, sampling_from_model:bool=False,
                 optimizer_type : OptimizerType = OptimizerType.EBM, optimizer : ModelBasedCounterOptimizer=None) -> np.ndarray:
    """
    Generate multiple counterfactuals that minimize the loss function using Bayesian Optimization.

    Parameters:
        query_instance:
            The instance to generate counterfactuals for.
        desired_class:
            The target class for the counterfactuals.
        adjacency_matrix:
            The adjacency matrix representing the causal structure.
        causal_order:
            The order of variables in the causal graph.
        proximity_weight:
            Weight for proximity loss component
        sparsity_weight:
            Weight for sparsity loss component
        plausibility_weight:
            Weight for plausibility loss component
        diversity_weight:
            Weight for diversity loss component
        bounds:
            The bounds for each feature to search over (dict with feature names as keys and tuple (min, max) as values).
        model:
            The predictive model used to predict class labels.
        features_order:
            order of features in query instance
        masked_features:
            masked features vector (features to skip)
        categorical_indicator:
            True at the index where the variable should be treated as categorical
        num_cfs:
            The number of counterfactual instances to generate.
        init_points:
            Number of initial points for Bayesian Optimization.
        n_iter:
            Number of iterations for Bayesian Optimization.
        X:
            training dataset to sample from.
        sampling_from_model:
            true if you want to generate samples from model after sampling from data and generating with relationship graph
        optimizer_type:
            type of optimizer used on model to generate counterfactuals. If you use OptimizerType.Custom, you need to provide optimizer instance
        optimizer:
            instance of optimizer (use for optimizer_type = OptimizerType.Custom)

    Returns
    -------
    np.ndarray:
        The generated counterfactuals that minimize the loss function.
    """
    cfs = []
    for _ in range(num_cfs):
        cf = _generate_single_cf(query_instance, desired_class, adjacency_matrix,
                                 causal_order, proximity_weight, sparsity_weight,
                                 plausibility_weight, diversity_weight,
                                 bounds, model, categorical_indicator, features_order, masked_features=masked_features,
                                 cfs=cfs, X=X, init_points=init_points, n_iter=n_iter,
                                 sampling_from_model=sampling_from_model, optimizer_type=optimizer_type,
                                 optimizer=optimizer, bayesian_causality=False)
        cfs.append(cf)

    return np.vstack(cfs)



import pandas as pd
import numpy as np
from pgmpy.estimators import HillClimbSearch, BIC, MaximumLikelihoodEstimator, BayesianEstimator
from pgmpy.models import DiscreteBayesianNetwork


def discretize_dataframe(df: pd.DataFrame, categorical_flags: list[bool], bins: int = 5) -> pd.DataFrame:
    """
    Discretize continuous features in a DataFrame based on categorical flags.

    Parameters:
    ----------
    df : pd.DataFrame
        Input DataFrame.
    categorical_flags : list[bool]
        List indicating whether each column is categorical (True) or continuous (False).
    bins : int
        Number of bins for discretization (default = 5).

    Returns:
    -------
    pd.DataFrame
        DataFrame with continuous features discretized into bins.
    """
    if len(categorical_flags) != df.shape[1]:
        raise ValueError("Length of categorical_flags must match number of columns in df.")

    result_df = df.copy()

    for col, is_cat in zip(df.columns, categorical_flags):
        if not is_cat:  # Continuous feature
            # Use pandas.cut to discretize into equal-width bins
            result_df[col] = pd.cut(df[col], bins=bins, labels=False)

    return result_df


import numpy as np

def discretize_ndarray(data: np.ndarray, categorical_flags: list[bool], bins: int = 5) -> np.ndarray:
    """
    Discretize continuous features in a NumPy ndarray based on categorical flags.

    Parameters
    ----------
    data : np.ndarray
        Input 2D array of shape (n_samples, n_features).
    categorical_flags : list[bool]
        List indicating whether each column is categorical (True) or continuous (False).
    bins : int
        Number of bins for discretization (default = 5).

    Returns
    -------
    np.ndarray
        Array with continuous features discretized into integer bins.
    """
    if len(categorical_flags) != data.shape[1]:
        raise ValueError("Length of categorical_flags must match number of columns in data.")

    result = data.copy().astype(object)  # Use object type to allow mixed types if needed

    for col_idx, is_cat in enumerate(categorical_flags):
        if not is_cat:  # Continuous feature
            col = data[:, col_idx]
            # Compute bin edges and assign bin indices
            bin_edges = np.linspace(np.min(col), np.max(col), bins + 1)
            result[:, col_idx] = np.digitize(col, bin_edges[1:-1], right=True)

    return result


def train_bayesian_model(df: pd.DataFrame, categorical_flags: list[bool], bins: int = 5) -> DiscreteBayesianNetwork:
    def fit_dag_structure(discretized_df):
        hc = HillClimbSearch(discretized_df)
        best_model = hc.estimate(scoring_method=BIC(discretized_df))
        best_model.fit(discretized_df, estimator=MaximumLikelihoodEstimator)
        return list(best_model.nodes()), list(best_model.edges())

    discretized_df = discretize_dataframe(df, categorical_flags, bins)
    nodes, edges = fit_dag_structure(discretized_df)
    model = DiscreteBayesianNetwork()
    model.add_nodes_from(nodes)
    model.add_edges_from(edges)
    model.fit(discretized_df, estimator=BayesianEstimator)
    return model


import numpy as np
import pandas as pd

def generate_sample_from_csf(adjacency_matrix, csf, sample_order=None, categorical=None):
    num_vars = adjacency_matrix.shape[0]
    samples = np.zeros((1, num_vars))

    if categorical is None:
        categorical = [False] * num_vars  # Default to all variables being non-categorical

    if sample_order is None:
        sample_order = range(num_vars)

    # Assign noise values to variables without parents

    for j in range(num_vars):
        if np.all(adjacency_matrix[j, :] == 0):  # If no parents, assign noise
            samples[0, j] = csf[j]
            if categorical[j]:  # Round the value if the variable is categorical
                samples[0, j] = np.round(samples[0, j])

    # Generate samples based on the specified order
    for j in sample_order:
        parents = np.where(adjacency_matrix[j, :] != 0)[0]
        if len(parents) > 0:
            # Predicted values based on parents
            predicted_values = np.dot(samples[:, parents], adjacency_matrix[j, parents])
            samples[:, j] = predicted_values
            if categorical[j]:  # Round the value if the variable is categorical
                samples[:, j] = np.round(samples[:, j])

    return samples

def get_parents(adjacency_matrix):
    parents = []
    num_vars = adjacency_matrix.shape[0]
    for j in range(num_vars):
        if np.all(adjacency_matrix[j, :] == 0):
            parents.append(j)
    return parents


def generate_samples(adjacency_matrix, num_samples, bounds=None, sample_order=None, categorical=None):
    num_vars = adjacency_matrix.shape[0]
    samples = np.zeros((num_samples, num_vars))

    if categorical is None:
        categorical = [False] * num_vars  # Default to all variables being non-categorical

    if sample_order is None:
        sample_order = range(num_vars)

    # Assign noise values to variables without parents
    for i in range(num_samples):
        if bounds is None:
            noise = np.random.normal(size=num_vars)  # assign_csf
        else:
            noise = np.random.normal(size=num_vars)
        for j in range(num_vars):
            if np.all(adjacency_matrix[j, :] == 0):  # If no parents, assign noise
                samples[i, j] = noise[j]
                if categorical[j]:  # Round the value if the variable is categorical
                    samples[i, j] = np.round(samples[i, j])

    # Generate samples based on the specified order
    for j in sample_order:
        parents = np.where(adjacency_matrix[j, :] != 0)[0]
        if len(parents) > 0:
            # Predicted values based on parents
            predicted_values = np.dot(samples[:, parents], adjacency_matrix[j, parents])
            samples[:, j] = predicted_values
            if categorical[j]:  # Round the value if the variable is categorical
                samples[:, j] = np.round(samples[:, j])

    return samples



def print_cfs(qi, cfs, features):
    """
    Construct a DataFrame from a list of instances cfs,
    filling fields that are the same as in instance qi with '-'.

    Parameters:
    qi (dict): An instance represented as a dictionary.
    cfs (list of dicts): A list of instances, each represented as a dictionary.

    Returns:
    pd.DataFrame: Transformed DataFrame with matching fields replaced by '-'.
    """
    df = pd.DataFrame(cfs, columns=features)

    # Replace fields that are the same as in qi with '-'
    for col in df.columns:
        df[col] = df[col].apply(
            lambda x: '-' if x == qi[df.columns.get_loc(col)] else f'{qi[df.columns.get_loc(col)]}->{x}')

    return df


from typing import Sequence, Tuple, Dict, Optional, List, Self

import numpy as np
import pandas as pd
from interpret.glassbox import ExplainableBoostingClassifier
from overrides import overrides
from pgmpy.models import DiscreteBayesianNetwork

from .ACFX import ACFX
from .abstract import OptimizerType

class AcfxEBM(ACFX):
    """
        AcfxCustom: A Counterfactual Explanation Model (using EBM as blackbox)
    """
    def __init__(self, blackbox: ExplainableBoostingClassifier):
        """

        Parameters
        ----------
        blackbox:
            EBM blackbox explainer
        """
        super().__init__(blackbox)

    @overrides
    def fit(self, X:pd.DataFrame, pbounds:Dict[str, Tuple[float, float]], causal_order:Optional[Sequence[int]]=None,
            adjacency_matrix:Optional[np.ndarray]=None
            ,y=None, masked_features:Optional[List[str]] = None,
            categorical_indicator:Optional[List[bool]] =None, features_order:Optional[List[str]] =None,
            bayesian_causality:bool = False,  bayesian_model : Optional[DiscreteBayesianNetwork]=None,
            num_bins:Optional[int] = None) -> Self:
        self.optimizer_type = OptimizerType.EBM
        return super().fit(X, pbounds, causal_order,adjacency_matrix,
                    y, masked_features,categorical_indicator, features_order, bayesian_causality, bayesian_model, num_bins)




import numbers
from typing import List, Dict

import pandas as pd
import numpy as np
from overrides import overrides
from sklearn.linear_model._base import LinearClassifierMixin

from ..abstract import ModelBasedCounterOptimizer

class LogisticRegressionCounterOptimizer(ModelBasedCounterOptimizer):
    def __init__(self, model:LinearClassifierMixin, X: pd.DataFrame, feature_bounds:dict):
        if not hasattr(model, 'coef_'):
            raise AttributeError('model.coef_ must be set')
        if feature_bounds is None:
            raise ValueError("feature_bounds must be set")
        if not isinstance(feature_bounds, dict) or not all(
                isinstance(k, str) and
                isinstance(v, tuple) and
                len(v) == 2 and
                all(isinstance(i, numbers.Number) for i in v)
                for k, v in feature_bounds.items()
        ):
            raise AttributeError("feature_bounds must be a dict with string keys and tuple of two floats as values")

        self.__feature_bounds = feature_bounds
        self.model = model
        self.X = X

    @overrides
    def optimize_proba(self, target_class: int, feature_masked: List[str]) -> Dict[str, float]:
        optimized_instances = []

        for index, instance in self.X.iterrows():
            coefficients = self.model.coef_[target_class]  # Extract model coefficients for the target class
            direction = np.sign(coefficients)
            optimized_instance = instance.copy()

            for i, feature_name in enumerate(self.X.columns):
                if feature_name not in feature_masked:
                    continue
                if feature_name in self.__feature_bounds:
                    min_val, max_val = self.__feature_bounds[feature_name]
                    if direction[i] > 0:
                        optimized_instance[feature_name] = max_val  # Increase feature value if positive impact
                    else:
                        optimized_instance[feature_name] = min_val  # Decrease feature value if negative impact

            optimized_instances.append(optimized_instance)

        df_optimized = pd.DataFrame(optimized_instances)
        avg_optimized = df_optimized.mean().to_dict()
        return avg_optimized



from abc import ABC, abstractmethod
from typing import List, Dict


class ModelBasedCounterOptimizer(ABC):
    @abstractmethod
    def optimize_proba(self, target_class: int, feature_masked: List[str]) -> Dict[str, float]:
        """
        Modifies the instance to increase the probability of the target class by adjusting feature values.

        Parameters:
        -----------
        target_class:
            The desired class to optimize towards.
        feature_masked:
            List of interchangeable features

        Returns:
        -------
        Dictionary of feature names and their optimized values
        """
        pass
